<?php

/**
 * @file
 *  
 * Supports the node-specific APIs and hooks for relationship
 * Renders the relationship table form,  validates and saves links against nodes.
 *
 * A node can now have any number of 'statements' attached to it.
 * A 'statement' is an RDF 'triple' relating one resource with another.
 * EG,  the "next" resource from page one is page two.
 *
 * Simple and complex stuctures can be built from a collection of such
 * relationships.
 * 
 * Primary statements behave as if they are attached directly to the
 * subject - the node. They are exposed for editing in the node edit form
 * and should be saved alongside it. Secondary statements are inferred
 * from logical patterns,  and although they may get stored in the node
 * as $node->relationships during processing and rendering,
 * they should not be saved unless you want every logical relationship 
 * to become explicit.
 *
 * This file provides the actual relationship editing UI form, and not much
 * else. To this end it uses the range_selectors.inc and combobox.js
 * to provide the dynamic UI.
 *
 * @package coders
 * @author Dan Morrison http://coders.co.nz/
 */


// we need to filter out some of the predicates, both for security
// and to keep the list managable.
// These constants describe what we show

/**
 * Intrinsic values are those predicates that have a direct mapping with
 * Drupal internals. Need to control access to this group.
 * An intrinsic property is one that has a drupalEquiv mapping value set.
 */
define('RELATIONSHIP_FILTER_INTRINSIC_VALUES',1);

/**
 * Trivial values are predicates that have an importance/weight less than zero.
 * This includes predicates that are only ever implied. They should be not
 * show up on user edit forms
 */
define('RELATIONSHIP_FILTER_TRIVIAL_VALUES',2);

/** @defgroup admin 
 *  
 * Administrator Interface
 * Hooks and functions for administering node relationships.
 * @{
 */


/**
 * Implementation of hook_nodeapi().
 *
 * based on taxonomy_nodeapi()
 */
function relationship_nodeapi(&$node,  $op,  $teaser = NULL,  $page=NULL) {
  //drupal_set_message("Calling relationship_nodeapi('$op' ".$node->nid);
  
  $active_node_types = variable_get('relationship_nodes',array());
  if ($node->type || (! $active_node_types[$node->type])) {return;} // Not a node type we deal with
  
  switch ($op) {
    case 'form post':
      debug('running form post',2);
      return relationship_form($node);
      break;
    case 'validate':
      return relationship_resource_validate($node);
      break;
    case 'insert':
      relationship_resource_save($node);
      break;
    case 'update':
      relationship_resource_save($node);
      break;
    case 'delete':
      
      // See if any relationships imply that this deletion should have a 
      // further-reaching effect. If this item isVampyreKingOf something,
      // then destroying this means that should be destroyed too...
      // Generally speaking, hasPart would exhibit this behaviour, but take care...
      if ($statements = relationship_query($node->nid, NULL, 1) )
      {
        foreach ($statements as $statement)
        {
          $term_def = relationship_get_term($statement['predicate']);
          // TODO
          // Find a way to abstract this operation 
          // - aren't there other times we'd want to use predicate hooks?
          if ($term_def->characteristics && isset($term_def->CascadeDeletion ) )
          {
            drupal_set_message("Cascading this action '$op' to other node ".$statement['object']);
            node_delete(array('nid' => nid($statement['object']) ,  'confirm' => 1) );
          }
        }
      }
    
        // It seems that the aliases have already been deleted before this runs.
        // How am I supposed to clean this up?
        // statements made using the alias as the subject will be left behind.
      //  foreach( relationship_get_aliases($node->nid) as $alias){
      //  relationship_resource_delete_statements($alias);
      //  }
      relationship_resource_delete_statements($node->nid);
      
      break;

    case 'load':
      $subject = ($node->nid) ? $node->nid : $_REQUEST['q'];
      debug("node load $subject",2);
      if(!$subject){return;}
      
      // Also extrapolate any useful built-in node data into statement form
      // This may seem a lot of work, but is used for meta tags and RDF
      debug("Analyzing built-in node properties of '$subject' in terms of statements",2);
      $intrinsic_node_statements = relationship_array2statements( (array)$node, $subject, relationship_get_property_map(CLASS_NODE) , TRUE, 'intrinsic property from node load' );
      
      // flag each of these statements as derived, so they don't get changed inappropriately
      foreach($intrinsic_node_statements as $i => $s){
        $intrinsic_node_statements[$i]['depth'] = RELATIONSHIP_IS_CALCULATED;
      }

      $statements = relationship_query($subject);
      
      // lookup alias properties also
      foreach(relationship_get_aliases($subject) as $alias){
        $statements = array_merge($statements,relationship_query($alias));
      }

      if ( $statements || $intrinsic_node_statements ) {
        $node->statements = array_merge($statements, $intrinsic_node_statements); 
        // drupal data takes precedence if there's any conflict
      }
      debug('Loaded node statements are:<pre>'.print_r($node->statements,1).'</pre>',3);

      break;

    case 'view':

      if ($page) {
        // Insert all our relationship links into the header,  to make the semantic web happen!
        if ( variable_get('relationship_add_statements_to_head',TRUE) ){
          relationship_add_statements_to_head($node->statements);
        }
      } // page?

    break; // case view

    case 'settings': 

      // Integration with Drupal core - extend node types with extra predicates.
      // Add a list of the available predicates
      // (properties) to the configure content type page
      return relationship_content_type_settings($node);
    
  } // switch
}

/**
 * Make a modification to a from.
 *
 * Note, this hook runs for every possible form!
 * Check to see if it's one we care about
 */
function relationship_form_alter($form_id, &$form) {
  # dsm(array($form_id,$form));
  if ( $form['#id'] == 'node-form') {
    $node_type = $form['type']['#value'];
    debug("Maybe adding relationship edit table to the edit form $form_id which edits nodes of type '$node_type' ",2);

    // check that we apply relationships to this type at all
    $active_node_types = variable_get('relationship_nodes',array());
    //drupal_set_message("relationships apply to ".print_r($active_node_types,1));
    if ((!$node_type) || (! $active_node_types[$node_type])) 
    {return;} // Not a node type we deal with

    $node = $form['#node']; 
    
    # dsm($node);
    // As form alter is now called to create the structure of the form before
    // we even look at the submitted data, it's hard to manipulate the form 
    // based on what's just happened. http://drupal.org/node/37194
    // I need to check the submission to see if any new fields have appeared (via client insertion)
    // To know how my new version of the form will look.
    // relationship_read_form does this, so we'll call it now.
    relationship_read_form($node);
    
    $form['relationship'] = relationship_form($node);
    // the final return is not needed because $form is a reference.
    
    // Enhance the node form a little with some more context
    $form['relationship']['#collapsible'] = TRUE;
    $form['relationship']['#collapsed'] = TRUE;

    if( variable_get('relationship_show_intrinsic_terms',FALSE) ){
      $form['relationship']['warning'] = array('#value' => t('<p><small>You are viewing all known intrinsic properties of the node, as well as the explicitly stated relationships. Making changes directly to some of these <i>may</i> interact strangely with other modules.</small></p>') );
    }
  }

  /**
   * Hook  into the CCK content type definitions, so we can configure
   *  applicable  properties here too
   */
  if($form_id == '_content_admin_type_edit'){
   
    $form['relationships'] = array(
      '#type' => 'fieldset',
      '#title' => 'Relationships',
      '#value' => t("The relationships that may apply to this content type."),
    );
    $subject_uri = NODE_TYPE_PREFIX.$form['type_name']['#value'];
    $resource =  relationship_get_resource_by_name($subject_uri,NULL,TRUE);
    
    // The relationship form here can only set two useful properties:
    // which predicates and which ontologies this content type will be able to use.
    // TODO replace this setting with a class canhaveproperty lookup
    $predicate_list = array( PREDICATE_CAN_USE_ONTOLOGY, PREDICATE_CAN_HAVE_PROPERTY, PREDICATE_SUBCLASS_OF, PREDICATE_SAME_AS );
    $predicates = array_map(relationship_get_term, $predicate_list);

    $form['relationships']['table'] = relationship_form($resource, $predicates);
    // Need to unshift this onto the FRONT of the submit queue, as the main submit aborts with a goto
    $form['#submit'] = array_merge(array('relationship_content_type_form_submit' => array()), (array) $form['#submit'] );    
  }
}


/**
 * Generate a form for displaying and selecting relationship applying to a node.
 *
 * The form is supported by some DHTML on the client.
 * @see coders_widgets.module
 *
 * @param $node object - the resource
 * @param $predicates optional list of predicates to be used in the relationship drop-down
 * @param $objects optional list of urls to be used in the relationship object drop-down
 *
 * @return HTML table form for adding metadata statements
 *
 * @TODO significant rewrite. Deal with nids 
 */
function relationship_form(&$resource, $predicates=NULL, $objects=NULL) {

  debug("Constructing relationship form",2);
  // Enable combo box functionality
  if(module_exist('coders_widgets')){
    coders_widgets_init('combobox');
    coders_widgets_init('paired_selects');
    coders_widgets_init('editable_hidden_fields');
  } else {
    drupal_set_message("This form requires coders_widgets.module to be enabled to do its stuff nicely.");
  }

  if (!$resource->statements) { 
    if ($resource->nid) {
      // fetch existing statements
      // Should have already been loaded, but..
      debug("shouldn't get here, rendering statements for a node that should have been initialized but wasn't");
      $resource->statements = relationship_query($resource->nid);
    }
    else {
      // new, empty resource
      if(! $resource){ $resource=(object)$resource; }
      $resource->statements = array();
    }
  }

  # drupal_set_message("Displaying relationship statements <pre>".print_r($resource, 1) ."</pre>");

  // List all existing data in a table here
  // There are a few extra ids and things in the layout
  // for javascript widgets to do their job.
  
  $rows = array(
    '#type' => 'formtable',
    '#id'   => 'metadata_entry_table',
  );
  $rows['header'] = array('#type' => 'tr', 'this', 'relates to', '', array('#value'=>'that','#width'=>'100%'), '');

  $rows['header'] = array('#value' => '<thead><tr><th>this</th><th>relates to</th><th></th><th width="100%">that</th><th></th></tr></thead>');

  $relationship_terms = relationship_get_terms();
  
  $show_trivial = variable_get('relationship_show_advanced_terms',FALSE);
  $show_intrinsic = variable_get('relationship_show_intrinsic_terms',FALSE);
  $can_edit_intrinsic = user_access('edit intrinsic properties');

  if(is_array($resource->statements)){
    relationship_get_titles($resource->statements);

    foreach ($resource->statements as $i => $statement) {
      // under 4.7 it seems I have to number all my multiple fields, can't just use [] no more
      $delta = check_name($i);

      debug("(possibly) Creating form row for statement $i",2);

      // TODO this routine needs to be refactored - it was constructed sorta ad-hoc
      // logic is currently broken down into the tiniest pieces, to really lay out whats happening
      $table_cell = array('#type' => 'td');
      $row = array(
        '#type' => 'tr',
        'subject_td'   => $table_cell,
        'predicate_td' => $table_cell,
        'change_td'    => $table_cell,
        'object_td'    => $table_cell,
        'action_td'    => $table_cell,
      );      


      $is_trivial = ($statement['depth'] < 0);
      $is_explicit = ($statement['depth'] == 0);
      $is_deduced = ($statement['depth'] > 0);

      if( $is_trivial && ! $show_trivial){continue;}


      // This statement should be deleted, remember this through the previews
      if($statement['delete']){
        debug('remember to delete this '.$statement['predicate'].' statement',3);
        $row['#class'] = 'delete';
        $row['#attributes']['style'] = 'color:#CCCCCC; text-decoration:line-through';
        $row['#attributes']['title'] = 'this statement will be deleted on save';
        $row[] = array(
          '#type' => 'hidden',
          '#value' => 'delete',
          '#parents' => array('statements',$delta,'delete'),
        );
        $is_explicit = FALSE; // disable further editing
      }

      // TODO One day I should put fine-tuned authentication here.
      // It's reasonable to set permissions on a predicate-by-predicate basis.
      
      $term_def = $relationship_terms[$statement['predicate']];
      $is_intrinsic = $term_def->drupalEquivalent;

      if( $is_intrinsic && ! $show_intrinsic){continue;}

      $can_edit = TRUE;
      if( $is_intrinsic && ! $can_edit_intrinsic ){ $can_edit = FALSE;}
      if( $is_deduced ){ $can_edit = FALSE; }

      // if a predicate has no defined range (eg 'Node' or 'Resource')
      // treat it as a literal ... UNLESS it already looks like an URL
      // these are just guesses to make up for incomplete ontologies.
      // Having no defined range is bad.
      $is_literal = (($term_def->range == CLASS_LITERAL) || (!$term_def->range && !is_uri($statement['object']) && !is_numeric($statement['object'])) );


      // Display subject only if it's different than expected (an alias)
      $row['subject_td'][] = ($resource->nid && ($statement['subject'] != $resource->nid))
                      ?  "<span class='subject'>{$statement['subject']}</span>" 
                      : '';

      $row['predicate_td'][] = array(
        '#type' => 'markup',
        '#value' => '<span class="predicate" title="'.$term_def->name."\n".check_plain($term_def->description).'">'. (($term_def)? $term_def->label : $statement['predicate']) . '</span>'
      );

      if($can_edit){
        // Statement is partially or fully editable
        
        // need hidden fields
        $row['subject_td'][] = array(
          '#type' => 'hidden',
          '#value' => $statement['subject'],
          '#parents' => array('statements',$delta,'subject'),
        );

        $row['predicate_td']['predicates'][] = array(
          '#type' => 'hidden',
          '#value' => $statement['predicate'],
          '#parents' => array('statements',$delta,'predicate'),
        );

        $row['object_td']['depths'][] = array(
          '#type' => 'hidden',
          '#value' => $statement['depth'],
          '#parents' => array('statements',$delta,'depth'),
        );

				// The HTML id of a changeable form element - used by widget code
        $target_object_id = "hidden_object_${delta}";
        $object_element = array(
          '#type' => 'hidden',
          
          // Cannot set my own ID anymore in 4.7?
          //'#attributes' => array('id' => 'hidden_object_'.$delta),
          // ah - it's no longer an attribute' 
          '#id' => $target_object_id,
          '#attributes' => array('class'=>'changeable'),
          '#value' => $statement['object'],
          '#parents' => array('statements',$delta,'object'),
        );

        $link_atts = array('id' => 'exposed_object_'.$delta);
        
        // can rename the title
        if ($is_explicit ){
          if(!$is_literal){
            // Only explicit links can be titled
            // Only save titles that are different from the default. 
            // No need to fill up the info with redundancies, and they have a disturbing habit
            // of getting out of date.
            $title = ($statement['title'] != $statement['object'])?$statement['title']:'';

            $title_element = array(
              '#type' => 'hidden',
              '#attributes' => array('id' => 'hidden_'.$delta),
              '#class' => 'changeable' , 
              '#title' => 'Label:',
              '#value' => $title,
              '#parents' => array('statements',$delta,'title'),
            );

            $row['object_td']['titles'][$delta] = $title_element;

            $link_atts['id'] = 'exposed_object_'.$delta;
          } else {
            $object_element['#class'] = 'changeable';
            $object_element['#title'] = 'Value:';
          }
        }

        $row['object_td']['objects'][] = $object_element;
#debug_pre($statement);
        $row['object_td'][] = relationship_make_link( $statement, $link_atts );

        
        if ($is_explicit && (! $is_literal)  && module_exist('coders_widgets') ){
          // Trigger possible widget script
          $row['change_td'][] = '<input type="button" value=">" '.
            ' onClick="change_select_range_to(\''.$target_object_id.'\',\''.$statement['predicate'].'\'); init_combo_box_named(\''.$target_object_id.'\') ; remove_element_named(\'exposed_object_'.$delta.'\') " title="Choose from other valid values for this property" />';
        }


        if ($is_explicit && module_exist('coders_widgets') ){
          // Only Explicit statements can be deleted
          // Writing button by hand as the drupal render is too bulky to fit in a table
          // This primarily works as javascript, but falls back as a submit action on noscript
          $row['action_td'][] = '<input type="submit" value="Del" name="edit[statements]['.$delta.'][delete]" '.
            ' onClick="if (typeof(delete_row_from_edit_table)==\'function\') {return delete_row_from_edit_table(this,this.name);}" />';
        } else {
          if ($is_trivial ) {
              $row['action_td'][] = "<small title='This actual statement is derived from data stored internally or elsewhere '>implicit</small>";
          }
        }

      } else {
        // can't edit this
        $row['object_td'][] = relationship_make_link( $statement );

        if( $is_deduced ){
          $row['action_td'][] = "<small title='This actual statement is in another page,  or due to the logical structure of the documents. It cannot be directly edited from this page '>... by implication</small>";

        }
        if( $is_intrinsic ){
          $row['action_td'][] = "<small title='This value is handled by the Drupal internals, or another module.'>built-in</small>";
        }
          // And can't do much more than that with this statement
      }
      
      $rows[] = $row;
     
    } // each statement
  } // has existing statements


  // Done rendering all known statements
  // now,
  // Append a new editable row for a new statement 

  // Initialize javascript selectbox updater
  include_once('range_selectors.inc');

  // Figure what predicates are appropriate to put into the select,
  $options_filter = 0;
  if(! ( $show_intrinsic && $can_edit_intrinsic )){
    $options_filter = RELATIONSHIP_FILTER_INTRINSIC_VALUES;
  }
  
  #debug_pre($predicates);
  
  if(! $predicates){
    // This lookup should fetch not just explicitly defined
    // properties, but also others according to the lookup rules in 
    // relationship_lookup_CanHaveProperty() eg, properties inheirited by virtue of class.
    // todo - relationship_lookup_CanHaveProperty() needs work

    $resource_type = ($resource->nid)? NODE_TYPE_PREFIX.$resource->type : $resource->type;
    
    debug("Retrieving all predicate relationships this resource type ({$resource_type}) MAY possible have",3);

    $predicate_list = relationship_query($resource_type, PREDICATE_CAN_HAVE_PROPERTY, NULL, 'object');
    $predicates = array_map(relationship_get_term, $predicate_list);

    debug('Predicates which are appropriate to display on this '.$resource->type.' form are:<pre>'.print_r(array_values($predicate_list),1).'</pre>' ,1);
    # debug_pre($predicates);
  
    $options_filter = ($options_filter | RELATIONSHIP_FILTER_TRIVIAL_VALUES);
  } else {
    debug('This form has been handed a list of which predicates to expose: '.join(', ',array_keys($predicates)) ,1);
  }


  // flatten the definitions into a (restricted) option list
  $predicate_list = relationship_terms_as_options($predicates,$options_filter);

  #debug('Predicates appropriate to display on this '.$resource->type.' form are:<pre>'.print_r($predicates,1).'</pre>' );
  debug('FILTERED Predicates appropriate to display on this '.$resource->type.' form, according to rules and inheiritance are:<pre>'.print_r($predicate_list,1).'</pre>' ,1);
  if(! $predicate_list){
    // This can only happen with a broken install or bootstrap.
    drupal_set_message(t("There are no known predicates (relationships or properties) in the local schema that are appropriate to apply to this resource. Create or import some predicate terms, then apply them to items of this type. The relationship selectbox will be empty."),'error'); 
  }

  $predicate_list = array_merge(array(''=>'...'), $predicate_list ); //insert blank first field 
  // didja know that a blank option is apparently invalid XHTML?

  // Inline a buch of javascript configuration script
  $rows[]=array('#value' => range_selectors($predicates, 'predicates][', 'objects][' ) );

  // Don't put anything in the selector yet?
  $object_list = array('- Choose a relationship first'); 
  
  // the new row is constructed here  
  $rows[] = _relationship_statement_fields( $predicate_list, $object_list, $resource->uri );
  
  $form = array($rows);

  $group =  array(
    '#type'   => 'fieldset',
    '#title'  => t('Relationships and Metadata'),
  );
  $group['relationships']= $form;
  
  $group['relationships']['help']['#value'] = t("
      <p>For more information about the relationships available to you, and their possible values, 
      See %help_link</p>
    ", array('%help_link' => l(t('Relationship Help (Predicates)'),'admin/help/relationship/predicates') )
  );

  if(module_exist('forms_api_backport')){
    // in 4.7 emulation mode
    debug('Need to find the right calls to backport just PART of a form');
    return drupal_get_form('statement_table', $group);
  }

  return $group;
}



/**
 * Convert a list of term def objects into an array
 * suitable for using in a select box.
 *
 * They will be indexed either of shortname (name) or longname (uri) depending on global settings
 *
 * Checks that only resources of type 'property' show up,
 * as relationship_lookup_CanHaveProperty() is currently a bit inaccurate
 * 
 * @param The current list or subset of predicate terms definitions.
 * @param $filter A boolean combination of options indicating which of the
 *        available terms NOT to show. 
 *        By default both filters are on - you must choose to turn them off.
 * @see RELATIONSHIP_FILTER_INTRINSIC_VALUES
 * @see RELATIONSHIP_FILTER_TRIVIAL_VALUES
 * @see drupal_get_terms()
 */
function relationship_terms_as_options($terms , $filter = 3 ) {
  if(! is_array($terms)){
    trigger_error("Invalid arg '$terms' to relationship_terms_as_options() ",E_USER_WARNING);
    return array();
  }
  $options = array();

  // index this list on long or short names
  $relationship_long_names = variable_get('relationship_long_names', FALSE);
  
  foreach($terms as $i => $term){
    debug("Filtering? $term->name weighted $term->weight , built-in? $term->drupalEquivalent ",3);
    // Skip the low-level terms - they are too trivial to list
    if(( $filter & RELATIONSHIP_FILTER_TRIVIAL_VALUES ) && ($term->weight < 0)){ continue; }

    // Do not list built-in properties unless given special permission
    if(( $filter & RELATIONSHIP_FILTER_INTRINSIC_VALUES ) && ($term->drupalEquivalent)){ continue; }

    if($term->type != CLASS_PROPERTY ){ continue; }

    // @TODO greater granularity of access can be applied here
    // else
    $indexid = ($relationship_long_names)?$term->uri:$term->name;
    $options[$indexid] = $term->label;
  }
  return $options;
}



/**
 * Ensure a sting is able to be used as an XML, CSS or Javascript ID.
 * Basically strip out all non-alpha-numerics
 * http://www.w3.org/TR/REC-xml/#NT-Name
 * @see form_clean_id() - which should have done this
 */
function check_name($name){
  return preg_replace('|[^a-zA-Z0-9_]+|','_',$name);
}


/**
 * Read an edit form into a set of statements.
 *
 * Validating a nodes statements means checking the 'statements' in the node
 * are legal.
 * Nodeapi callback
 *
 * @param $node Expected to contain
 *              An array of field values, as submitted by an edit form.
 *              Expected to contain arrays subjects[] , predicates[] objects[]
 *              and optionally titles[]. 
 *              Horizontal rows in these vertical columns are our statements
 * @return node object
 */
function relationship_resource_validate(&$node) {
  debug('Validating form input '.print_r($node, 1),3 );
  
  // this should have been done already by form_alter ?.
  relationship_read_form($node);

  // If given, also allow relationship statements to be absorbed over top of
  // intrinsic node properties. This gives low-level access to the guts of a
  // node (probably a bad thing)
  $data_cloud = relationship_statements2array($node->statements );
  foreach($data_cloud as $predicate => $object){
    if( isset($node->$predicate) &&  ( $node->$predicate != $object) ){

      // Filter for user-access restrictions. Don't want to expose Drupal guts randomly
      if( user_access('edit intrinsic properties') ){
        debug("Over-writing node $predicate = {$node->$predicate} with $object ",2);
        $node->$predicate = $object;
      } else {
        drupal_set_message(t('You don\'t have permissions to modify the %predicate directly on this resource, but thanks for trying!',array('%predicate'=>$predicate) ));
      }
    }
  }

  // init title if needed
  relationship_get_titles($node->statements, TRUE);
  return $node;
}

/**
 * Read form values and absorb them into the given resource object as
 * 'statements'
 * This bypasses the form API, as it may heve to find fields added on-the-fly at
 * the client side.
 * @param $resource - an object, passed by ref that may be updated
 * @param $subject string optional, gets set as the 'subject' of any statements
 * if the subject is otherwise undefined. Should be the $node->nid or the $term-
 * >identifier or whatever
 */
function relationship_read_form(&$resource,$subject=''){
  debug("Reading submitted form values back into statements. Posted values are:",2);
  debug_pre($_REQUEST['edit'],3);
  debug_pre($_REQUEST['edit']['statements'],2);

  // form api doesn't recognise any fields added by the client (for security reasons) 
  // but I need to bypass that . Read directly from the submission
  $submitted_statements = $_REQUEST['edit']['statements'];
  
  if (is_array($submitted_statements) ) {
    foreach ($submitted_statements as $ix => $statement) {

      if ($statement && $statement['predicate'] && $statement['object']) {
        $statement['subject'] = ($statement['subject'])? $statement['subject'] : $subject;
        $statement['object']  = trim($statement['object']);
        $statement['title']   = trim($statement['title']);
        $statement['why']     = ($statement['why'])?$statement['why']:'user/'.$GLOBALS['user']->uid;
        // It's possible a 'sid' number or a 'discard' flag will be sent through here also. Keep it.

        $key = _relationship_hash_key($statement);
        if((string)$ix != $key){ // strange, comparing the number 0 with a string is always true. cast to fix
          // re-index this item
          unset($resource->statements[$ix]);
          #            unset($_REQUEST['edit']['statements'][$ix]);
        }
        $resource->statements[$key] = $statement;
      }
    }
  }
  debug("Finished reading form. statements are now:<pre>".print_r($resource->statements,1)."</pre>",2);
  return $resource;
}

/**
 * Save the relationship associations (statements) of a given resource.
 *
 * AKA relationship_node_save() if following drupal module conventions.
 *
 * The resource object handed in (usually a node) is expected to have an array
 * called 'statements' attached to it. Each of these statements are serialized
 * into the database.
 * @see statements for the structure of a 'statement'.
 *
 * If it wasn't explicitly set in the statement, 
 * the 'subject' is assumed to be this node.
 * Statements can have an optional title.
 *
 * $statements = array(
 *   array('predicate'=>'IndexedBy',  'object'=>'sitemap'),
 *   array('predicate'=>'LinksTo',  'object'=>'sitemap'),
 *   array('predicate'=>'LinksTo',  'object'=>'homepage', 'title' = 'Home'),
 *);
 *
 * As the method is currently 'wipe the old and replace the new'
 * any statement IDs are lost on each new save.
 * This is not a problem yet... still not actually using rsids.
 *
 * @param $resource An object containing an array of statements
 * @param $uri The subject identifier. Usually derived from the resource
 * properties itself. Used to fill in incomplete statements.
 * @return void
 */
function relationship_resource_save(&$resource,$uri='') {
  
  if ( (! $uri) && (! $uri = $resource->nid ) && (! $uri = $resource->uri ) ) {
    drupal_set_message("Can't save metadata with no node ID or URI"); return;
  }

  // careful not to wipe statements if the nodes data is simply undefined, 
  // maybe we didn't load the full set in the first place.
  if (is_array($resource->statements) ) {

    // We no longer clear and start again, each statement is added or deleted explicitly
    // relationship_resource_delete_statements($uri);
    debug("Saving statements",1);
    #debug_pre($resource->statements);
    #debug_pre($resource);

    foreach ($resource->statements as $statement) {
      if (is_array($statement) && $statement['predicate']) {
        if (! $statement['subject']) {
          $statement['subject'] = $uri; 
          debug("semi-anonymous statement (".$statement['predicate'].") has been submitted, guessing the subject was meant to be '$uri",2);
          // brand new nodes may not have had the nid available at statement time.
        }
        if($statement['delete']){
          drupal_set_message("Deleting statement about ".$statement['predicate']);
          relationship_delete_statement($statement);
        } else {
          relationship_save_statement($statement,'Saving Resource');
          // invalid statements (like a blank row from form submission) are discarded under here
        }
      } // valid
    } // each
  } // statements exist
  else
  {
    debug("Asked to save a resource ($uri) with no attached statements. This may be an error");
  }
}





/**
 * Generate form elements for defining relationships.
 *
 * subject (current node, implied from context)
 * predicate (relationship term)
 * object (freeform,  probably a node id or URL)
 *
 * Return as an array so the respective parts can be
 * dealt with separately (put into a table row)
 *
 * Helper for node_form
 *
 * @param $predicates optional list of predicates to be used in the relationship drop-down
 *                    if not defined, ALL things of type 'Property' will be listed
 * @param $objects optional list of titles=>uris to be used in the relationship object drop-down
 * @returns an array to be put into a table row.
 * @todo test noscript
 */
function _relationship_statement_fields($predicate_list=NULL, $object_list=NULL, $subject='') {

  // fudging the parents array fools form_api into constructing the right name, id & data structure.
  // This row must be set with a null value (it's always a new entry) or forms api re-populates it
  // by being clever, which we don't need.
  $predicate_selector = array(
    '#type' => 'select',
    '#target' => 'object_selector',
    '#title' => '',
    '#default_value' => '',
    '#value' => '',
    '#options' => $predicate_list,
    '#description' => NULL,
    '#parents' => array('statements','0','predicate'),
  );

  $object_selector = array(
    '#type' => 'select',
    '#id' => 'object_selector',
    '#title' => '',
    '#default_value' => '',
    '#value' => '',
    '#options' => $object_list,
    '#description' => NULL,
    '#attributes' => array('class' => 'combo'),
    '#parents' => array('statements','0','object'),
  );
  $hidden_subject = array(
    '#type' => 'hidden',
    '#id' => 'hidden_subject',
    '#default_value' => $subject,
    '#parents' => array('statements','0','subject'),
  );
 
 // Construct a row, rather laboriously
  $row = array(
    '#type' => 'tr' ,  
    'subject_td'   => array(
      '#type'        =>'td', 
      'data'         => array('#value' => 'this'),
      'subjects'     => array('',$hidden_subject) ,
    ),
    'predicate_td' => array(
      '#type'        =>'td', 
      '#attributes'  => array('colspan'=>'2'),
      'predicates'   => array('' => $predicate_selector ),
    ),
    'object_td'    => array(
      '#type'        => 'td',
      'objects'      => array('' =>  $object_selector ),
    ),
  );
  
  if(module_exist('coders_widgets')){
    // add DHTML form enhancements
    coders_widgets_init('dynamic_table_rows');
    $row['predicate_td']['predicates']['']['#type'] = 'paired_select';
    $row['action_td'] = array(
      '#type'  => 'td',
      'data'   =>'<input type="button" value="New" name="new_statement" id="new_button" '.
      'onClick="if (typeof(add_row_to_edit_table)==\'function\')'.
      '{return add_row_to_edit_table(this);}'.
      'else{alert(\'no dhtml function available\');return false;}" />',
    );
  }
   
  return $row;
}



/**
 * A query lookup callback, returns all predicates that may be available to a type.
 *
 * All the properties a subject can have!
 * This includes a logical lookup not implied by the ontology.
 * These callback lookups are where we CAN put business logic that we can't
 * yet express in statements.
 *
 * In addition to those properties added explicitly in the datastore, we should return 
 * - all properties from any associated ontologies, 
 * - all properties that may apply to the superclass (node)
 * - all wildcard properties that have NO domain defined.
 *
 * @TODO check that members of ontologies really are 'properties'
 *       I'm getting a few false properties due to classes, non-node properties
 *       and ontologies being inside ontologies too.
 *       I don't really want EVERYTHING in an ontology.
 * @ingroup ontologies
 */
function relationship_lookup_CanHaveProperty($subject){
  if(!$subject){$subject = CLASS_RESOURCE;}

  debug("What properties is a '$subject' allowed to have?",2);
  $statements = array();
  
  // Lookup possible predicate groupings - used when importing
  // content type definitions from custom ontology sets
  $possible_ontologies = relationship_query($subject,PREDICATE_CAN_USE_ONTOLOGY,NULL,'object');
  foreach($possible_ontologies as $ontology_name){
    // for all ontologies this class participates in...
    $ontology_statements = relationship_query(NULL, PREDICATE_IS_PART_OF, $ontology_name);
    // each member of that ontology
    foreach($ontology_statements as $statement){
      // Note - that was a reverse lookup, so we use the 'subject' as our result
      // adjust all the derived statements to refer to the current subject
      $s = relationship_statement( $subject, PREDICATE_CAN_HAVE_PROPERTY , relationship_get_shortname( $statement['subject'] ) );
      $s['depth'] = (int)$statement['depth'] + 1;
      $s['why'] = $statement['why']."; in ontology $ontology_name, which is associated with ".$subject;
      $statements[_relationship_hash_key($s)] = $s;
    }
  }
 
  // all properties that may apply to the superclasses (node,Resource)
  if($subject != CLASS_CLASS){ // avoid loop
  // CLASS_RESOURCE
    $superclasses = relationship_query($subject, PREDICATE_SUBCLASS_OF,NULL,'object');
    // if(! $superclasses){debug("odd, this $subject doesn't seem to be an anything - it has no superclass");}
    foreach($superclasses as $superclass){
      debug($subject." is a subclass of '$superclass', so it can have its properties also",2);

      // TODO should this be recursive?
      $superclass_statements = relationship_query($superclass, PREDICATE_CAN_HAVE_PROPERTY, NULL);
      // adjust all the derived statements to refer to the current subject
      foreach($superclass_statements as $statement){
        if(!$statement['object']){debug('Query returned a statement with no object. this is not right. This code should not have been reached');}

        $s = relationship_statement( $subject, PREDICATE_CAN_HAVE_PROPERTY , relationship_get_shortname($statement['object']) );
        $s['depth'] = (int)$statement['depth'] + 1;
        $s['why'] = $statement['why']."; in ontology $ontology_name, which is associated with class: ".superclass;
        $statements[_relationship_hash_key($s)] = $s;
      }
    }
  }

  debug("Properties a '$subject' may have by virtue of lookup or inheiritance: ".join(', ',array_keys($statements)) ,2);

  foreach(relationship_get_terms() as $term){
    // all wildcard properties that have NO domain defined.
    if($subject && $term->name && !$term->domain)
    {
      $s = relationship_statement( $subject, PREDICATE_CAN_HAVE_PROPERTY , $term->name );
      $s['depth'] =  RELATIONSHIP_IS_CALCULATED;
      $s['why'] = "$term->name is a predicate term with no domain limit";
      $statements[_relationship_hash_key($s)] = $s;
    } 
    else if($term->domain == CLASS_RESOURCE)
    { // This term is explicitly defined as applying to all resources 
      $s = relationship_statement( $subject, PREDICATE_CAN_HAVE_PROPERTY , $term->name );
      $s['depth'] =  RELATIONSHIP_IS_CALCULATED;
      $s['why'] = "$term->name is defined to be a property available to all resources";
      $statements[_relationship_hash_key($s)] = $s;
    }
  }

  $reasons = array_map(create_function('$s', 'return $s["why"];'), $statements);
  debug("Properties a $subject may have by virtue of lookup, inheiritance or wildcards (not counting direct statements): <li>".join("</li>\n<li>",array_keys($reasons))."</li>" ,2);

  return $statements;  
}


/**
 * Vanilla data dump of the statements about a given resource
 */
function relationship_list_statements($resource) {
  if(! is_array($resource->statements)){return '';}

  // May optionally be hiding the built-in data and only show the relationship.module additions
  $show_trivial = variable_get('relationship_show_advanced_terms',FALSE);
  $show_intrinsic = variable_get('relationship_show_intrinsic_terms',FALSE);
  $relationship_terms = relationship_get_terms();
  
  foreach ($resource->statements as $statement) {
    $pred = $statement['predicate'];

    if(($pred == 'title') || ($pred == 'label')){
      // if the resource has a label or title defined in this statement
      // collection, display that as the identifier instead
      $label = $statement['object'];
    }

    // Filter out the data bits we don't want to see.
    $is_trivial = ($statement['depth'] < 0);
    if( $is_trivial && ! $show_trivial){continue;}
    $term_def = $relationship_terms[$pred];
    $is_intrinsic = $term_def->drupalEquivalent;
    if( $is_intrinsic && ! $show_intrinsic){continue;}

    if(is_array($pred)){debug_pre($statement);}

    if($term_def->{PREDICATE_RANGE} && ($term_def->{PREDICATE_RANGE} != CLASS_LITERAL)){
      // This object can probably be linked to
      $object_link = l( $statement['object'], variable_get('relationship_browser_root', 'relationship/view') ,NULL, 'edit[subject]='.urlencode($statement['object'])  );
    } else {
      $object_link = '';
    }
  
    $output .= "<dt>".$pred."</dt>\n";
    $output .= "<dd>".($object_link?$object_link:$statement['object'])."</dd>\n";
    
  }
  if(!$output){$output = t('<p>No data available</p>');}
  else{$output = '<dl>'.$output.'</dl>';}
  return $output;
}



/**
 * List everything about everything. Page.
 *
 * Menu callback.
 * Admin function / diagnostic / scaffolding
 * Should really use pager
 */
function relationship_list_all() {

  drupal_set_title('Dumping all relationships in the database table');
  $output = 'Debug info only - this will get too big very quickly';
  
  $all_relationship_lookup = db_query('SELECT * FROM {relationship_statement} order by subject;');
  $rows = array();
  while ($statement = db_fetch_array($all_relationship_lookup) ) {
    $row = $statement;
    $row['subject'] = l( scrunch_url( $statement['subject'] ), 'relationship/view', NULL, 'edit[subject]='.urlencode($statement['subject']) );
    unset($row['title']);
    $rows[] = $row;
    $headers = array_keys($statement);
  }
  unset($headers['title']);
  $output .= theme('table', $headers  , $rows);

  print theme('page',$output);  
}


function relationship_choose_resource_form(){
  $form = array();
  $form['uri'] = array(
    '#type' => 'textfield',
  );
  $form['edit'] = array(
    '#type' => 'submit',
    '#value'=> 'Edit',
  );
  $form['view'] = array(
    '#type' => 'submit',
    '#value'=> t('View'),
  );

  return drupal_get_form('relationship_choose_resource_form', $form);
}

/**
 * Expose all statements about any subject.
 *
 * Menu callback.
 * This trats the given resource as a stand-alone collection of statements, with
 * no special meaning or treatment.
 *
 * The target subject can either be handed in as the remainder or the URL
 * or explicitly as a 'uri' form parameter.
 * This is a generic edit form for a resource. A resource is represented as
 * a collection of statements about the same thing.
 *
 * @see relationship_list_all()
 * @todo make this a saving edit form
 */
function relationship_data_page($op='View',$subject='') {

  $op = $_REQUEST['op'] ? $_REQUEST['op'] : $op ; // Form op overrides path op
  
  // absorb subject from any direction possible
	// I'll allow requests for both ?edit[subject]=thing and ?subject=thing
  // Or just appended to the request (for shortnames only)
  $edit = array_merge( $_REQUEST,($_REQUEST['edit'])?$_REQUEST['edit']:array());
  $subject = 
  	$subject ? $subject :
    	( $edit['uri'] ? $edit['uri'] : 
        ( $edit['identifier'] ? $edit['identifier'] : 
        	$edit['subject'] 
        )
      );
  
  $actions = array('View'=>'view','Edit'=>'edit','RDF'=>'rdf');
  foreach($actions as $label=>$command){
    $active = ($op==$label)?array('class'=>'active'):NULL;
    $actions[$label] = l( $label, 'relationship/'.$command, $active , 'edit[subject]='.urlencode($subject) );
  }
  $out .= theme_item_list($actions,NULL,'ul',array('class' => 'tabs primary'));


  if(!$subject){
    drupal_set_message(t('No subject defined. Need to call this page with an id or a URI as the <code>?uri=</code> argument'),'error');    
    #dsm($edit);
    print( theme('page', relationship_choose_resource_form() ));
    return;
  }
  
	$resource = relationship_get_resource_by_name($subject, NULL, TRUE);

  if(! $resource ){
    drupal_set_message("'$subject' is totally unknown to me. Creating a placeholder of type generic resource.");
    $resource = (object) array(
      PREDICATE_IDENTIFIER => $subject,
      PREDICATE_TYPE => CLASS_RESOURCE,
    );
  }

	if(! $resource->uri ){
	  // brand new
	  $resource->uri = $subject;
	}

  if($op == t('edit')){
    // load submission data (if any)
    // Note this merges existing info with submitted
    $resource = (object)array_merge((array)$resource,$edit);
    relationship_resource_validate($resource);
  }
  
  if($op == t('Submit')){
    drupal_set_message("Saving '$subject'");
    $resource = (object) relationship_statements2array( $edit['statements'], NULL, TRUE );

    if(! $resource->uri ){ // brand new
      $resource->uri = relationship_get_longname($subject);
    }
    relationship_resource_save($resource);
  }


  drupal_set_title("$op All info about '$subject'");


  // $out .= form_textfield( t('URI'), 'uri', $resource->uri ,80,200);

  $out .= t("
    Viewing the properties and relationships that apply to the 
    <b>%type</b> named '%name'. ",
    array(
      '%type' => $resource->type?$resource->type:CLASS_RESOURCE, 
      '%name' => relationship_get_display_label($resource) 
    )
  );

  if($op == t('edit')){
    $form = relationship_form($resource);
  
    $form['submit'] = array(
      '#type' =>   'submit',
      '#value'=>   t('Submit'),
    );
    $form['#redirect'] = FALSE;
    
    $out .=  drupal_get_form('relationship_resource_form', $form);
  } else {
    // Assume 'Viewing'
    $out .= relationship_list_statements($resource);
  }
  
  print( theme('page', $out ));
}


/*! 
 *  @}
 */


/**
 * Save the parameters set on the content type definition page
 * (CCK hook)
 * @see relationship_form_alter() for the other end of this process
 */
function relationship_content_type_form_submit($form_id,$edit){
  //dsm("save me! Ready to save the relationship values");  
  $content_def = (object)$edit;
  $subject_uri = NODE_TYPE_PREFIX.$content_def->type_name;
  $content_def = relationship_read_form($content_def,$subject_uri);
  //dsm($content_def);  

  // About here, we ensure that this resource is of class node definition
  $type_class_statement = relationship_statement($subject_uri,PREDICATE_TYPE,CLASS_NODE);
  $content_def->statements[] = $type_class_statement;

  relationship_resource_save($content_def,$subject_uri);

}

?>
