<?php

/**
 * @file RDF datastore
 * @author Dan Morrison http://coders.co.nz/
 *
 * Interface to the database.
 *
 * Retains simple statements and returns simple queries.
 * No inferencing or logic at all.
 * No built-in support for aliases even.
 *
 * This implimentation of a datastore just demonstrates the interface
 * expected. 
 * Use a copy of it as a stub for requests on other datastore implimentations.
 *
 */
 
 
Class relationship_datastore{
  
  function relationship_datastore() {
    debug("New DataStore initialized. Non-caching, non inferencing direct link with the database table. This engine natively supports: ". implode(', ', $this->supported_characteristics() ), 1);
  }
  
  var $supported_characteristics = array('inverseOf');
  /**
   * Return the predicate characteristics that this datastore supports.
   *
   * To support swapping differnt datastores in and out, 
   * A call to this function will return a list of the subset of OWL
   * that is handled internally.
   * The returned array may be as simple as ('sameAs' and 'inverseOf')
   * or it may do all sorts of logic.
   */
   
  function supported_characteristics($characteristic=NULL){
    return ($characteristic)? $this->supported_characteristics[$characteristic] : $this->supported_characteristics;
  }
  
  /**
   * Insert a statement into the store.
   *
   * Optional parameters describe the justification for inserting the
   * statement, such as the provenance (who sez?) or the prepositions
   * that led to this inferred statement. This may be used for an audit
   * trail.
   * 
   * @return whether the statement add was successful. 
   *         Duplicate statements are NOT added.
   */
  function add($statement, $message=NULL, $previous_ref=NULL)
  {
    $key = _relationship_hash_key($statement);

    debug("DataStore adding [$key]" ,2);

    if($this->query($statement)){
      debug("Already knew that [$key]. Not saving" ,2);
      return FALSE; // already knew that
    }

    if($statement['depth'] < 0 ){
      // statements with negative depth are trivial, temporary, transient, dynamic or calculated values
      debug("Trivial, [$key]. Not saving" ,2);
      return FALSE; // Not saving trivial values in the database ... yet
    }


    // This is a very public API - lets be paranoid about input
    if(! relationship_validate_statement($statement)){
      debug('Invalid statement, not adding to datastore'.$key ,0 );
      return FALSE;
    }
    if(!$message){ $message = $statement['why']; }
    if(!$message){ $message = "no reason"; }

    db_query('INSERT INTO {relationship_statement} (subject,  predicate,  object,  title, depth) VALUES ("%s", "%s",  "%s", "%s", %d)',  $statement['subject'],  $statement['predicate'],  $statement['object'],  $statement['title'],  (int)$statement['depth']);
    
    debug("DataStore stored new fact <b>$key</b> because $message" . ($previous_ref?" and "._relationship_hash_key($previous_ref) : "") ,2);
    return TRUE;
  }


  /**
   * Insert a collection of statements into the store.
   *
   * @return number of statements successfully added.
   * @see add_array()
   */

  function add_array($statements, $message='', $previous_ref=NULL) {
    $count = 0;
    debug('Adding '.count($statements).' statements '.$message,3);
    
    if(is_array($statements)){
      foreach($statements as $statement){
        if($this->add($statement, $message, $previous_ref)){$count ++;}
      }
    }
    return $count;
  }


  /**
   * Query the datastore for statements that could answer the given question
   * ?subject ?predicate ?object
   * Any or all of the parameters may be null.
   *
   * It can also be called with the three parameters packed into the first - 
   * A query is just a statement with gaps in!
   *
   * @subject If this is a full statement, the query is constructed so as 
   *          to fill in the undefined columns.
   * @return an array. An empty array should be taken as a valid response
   *              meaning there is no data. FALSE, OTOH, means I don't
   *              know.
   */
  function query($subject=NULL,$predicate=NULL,$object=NULL) {

    if(is_array($subject)){
      $statement =  $subject;
    } else {
      $statement = relationship_statement($subject,$predicate,$object);
    }

    debug("Querying Database about {".implode(", ",$statement)."}" , 3);

    $clause = $this->_clause_from_statement($statement);

    $result = db_query('SELECT * FROM {relationship_statement} '.$clause.' ORDER BY predicate,  object ' ,  $subject);

    $subject_statements = array();
    while ($statement = db_fetch_array($result) ) {
      // Statements are indexed on hash keys to avoid repetition.
      // Dupes are squashed here
      if(! $statement['why']){$statement['why'] = 'directly from database';}
      $subject_statements[ _relationship_hash_key($statement) ] = $statement;
    } // retrieved all first-level statements

    return $subject_statements;
  }

  /**
   * Delete all statements matching the given criteria
   * If given an actual statement array as the first parameter, that statement,
   * or statements matching the given statement pattern are deleted.
   */
  function delete($subject=NULL,$predicate=NULL,$object=NULL) {
    if(is_array($subject)){
      $statement =  $subject;
    } else {
      $statement = relationship_statement($subject,$predicate,$object);
    }

    $clause = $this->_clause_from_statement($statement);
    if(! $clause){
      drupal_set_message("Warning! Almost deleted all statements just then. datastore->delete was called with no restriction clause, this was probably incorrect. If you really want to truncate the table, use the database admin instead.");
      return FALSE;
    }

    debug("<h4>Deleting statements matching query {".implode(", ",$statement)."} $clause </h4>" , 2);

    return db_query('DELETE FROM {relationship_statement} '.$clause);
  }

  /**
   * Convert a statement into a SQL WHERE query
   * Helper for db queries
   */
  function _clause_from_statement($statement) {
    // Generate ad-hoc query
    $clauses = array();
    $columns = array('subject','predicate','object','title','sid');
    foreach($statement as $key=>$val){
      if($val && in_array($key,$columns) ){ // should I use isset?
        $clauses[] = " ($key = '". str_replace('%', '%%', db_escape_string($val)) ."') "; 
      }
    }
    $clause = implode(' AND ',$clauses); 
    if ($clause){ $clause = ' WHERE ' . $clause; }
    return $clause;    
  }

  /**
   * stub
   */
   function get_cannonic($subject,$predicate,$max_depth=0) {
     warn("Stub called");
   }


  /**
   * stub
   */
  function get_all($subject,$predicate,$max_depth=0) {
     warn("Stub called");
  }
  
  /**
   * stub
   */
  function get_by_key($key) {
     warn("Stub called");
  }

  /*
   * Return all statements discovered since last requested.
   * flushes the internal list
   * stub
   */
  function get_new()
  {
    warn("Stub called");
    $new_facts = $this->new_facts;

    // no longer new.    
    $this->new_facts = array();

    return $new_facts;
  }

}

