<?php

/*
 * Dear coder
 *
 * If you are looking at this code, It's probably best to start by having a look
 * at docs.php or the autogenerated API docs. Theres a conceptual overview over there.
 * .dan.
 */

/**
 * @file relationship api
 * @author Dan Morrison <dan [AT] coders.co.nz>
 * 
 * Interface for modules and other code to set and get relationship statements.
 *
 * This layer takes care of the datastore abstractions, and
 * provides a set of utility accessor functions to establish and query
 * relationships between resources, defined as subject-predicate-object statements.
 *
 * No rendering or even forms or settings in this file, only datastore abstraction
 * and accessor functions.
 * 
 * As an API, it does not rely on functions other than those specifically
 * included (the datastore library).
 */

/**
 * @ingroup relationship_api
 * @{
 */

/**
 * For hashing shorthand statement indexes
 * @see _relationship_hash_key()
 */
define('STATEMENT_SEPARATOR','-');

define('RELATIONSHIP_TERMS_CACHE_KEY','relationship_terms');
define('RELATIONSHIP_PROPERTY_MAP_CACHE_KEY', 'relationship_property_maps');

/*
 * Statement cache object
 * @todo make this more pluggable
 * The idea is that different datastores may use different databses 
 * (or even server queries) in different ways.
 * Thus, all my 'set' and 'queries' are interpreted by a black-box.
 */
include_once('rdf_datastore.inc');


/////////////////////////////////////////////////////////
// Programmers API
/////////////////////////////////////////////////////////
// Accessor Functions
//
//  Set,  get,  lookup terms and relations
//


///////////////////////////////////////
// Constructors & converters


/**
 * Just a helpful struct array constructor
 *
 * About the 'statement' model used internally:
 *
 * To retrieve the metadata for a node:
 *
 *     $statements = relationship_query($nid);
 *
 * Returns an array of statements.
 * Each statement may look something like:
 *
 *     $statement = Array 
 *         (
 *             [subject] => 198            // The current node
 *             [predicate] => isPartOf     // The relationship identifier
 *             [object] => 197             // The target node OR ALIAS
 *             [title] => Chapter 1        // The display title,  extracted from the target node itself
 *             [object_nid] => 197         // The target node ID,  useful if the defined 'object' was an alias path
 *        )
 *
 * Optionally,  it may also have a [depth] value set,  which indicates
 * how much logical recursion it took to come up with this statement.
 * A depth > 0 indicated this statement is IMPLIED,  not explicitly stated.
 * A depth < 0 is trivial, used for system statements
 * like a Drupal Alias is interpreted as a RDF:sameAs relationship
 * and other Drupal-managed data, like lastModifiedDate which may sometimes
 * be represented as statements, but is not expected to be serialized as such.
 *
 * Another optional property is
 *             [path] => node/197          // The link to display to this target
 * this gets calculated if needed when relationship_get_title() validates
 * the statement.
 *
 * The 'title' of a relationship may be defined at create or edit time 
 * (in which case it is stored in the database as the title field and doesn't change much)
 * or,  if left blank,  it can be retrieved each time it's used
 * (which allows it to reflect changes in the title of the target,  but is
 * probably inefficient in most cases,  as it causes a node load for all references).
 * Storing title in the triple table is an extension to the RDF model used internally,  
 * and is sort of experimental.
 *
 * @see relationship_save_statement() The serializer
 * @see relationship_set() Construct-and-save combined
 * @see relationship_array2statements() From turtle-like structure to triple structure
 *
 * @param $subject URI or node ID
 * @param $predicate shortname (or URI)
 * @param $object URI or Node ID
 * @param $title optional visible label for exposing this link
 * @param $why optional Something that explains why this statement was made, 
 *             eg an annotation by a user (uid), a deduced logic path (todo), or a system-defined fact (internal)
 * @return $statement assoc array structure.
 *
 */
function relationship_statement($subject,  $predicate,  $object,  $title=NULL,  $depth=NULL, $why=NULL) {
  $statement = array(
    'subject'     => (string)$subject,
    'predicate'   => (string)$predicate,
    'object'      => (string)$object,
    'title'       => (string)$title,
    'depth'       => (int)$depth,
 );
 if($why){ $statement['why']=$why; }
 return $statement;
}


/** 
 * Return just the named column out of a collection of statements.
 *
 * A vertical slice from a horizontal, 2D array.
 * @todo this may be more elegant, but is it more efficient than 
 * using trivial named accessor functions like relationship_get_object() ?
 */
function relationship_get_column($statements,$column) {
  if(! $column){return $statements;}
  
  return array_map(
    create_function('$statement', 'return $statement["'.$column.'"];') ,
    $statements
  );
}
/**
 * Return a named array, pairing the contents of column a as the key for
 * the contents of column b.
 *
 * There may be a php function I'm missing that does this better.
 */
function relationship_get_assoc_column($statements, $columna, $columnb, $as_shortnames) {
  $a = array();
  foreach($statements as $s){
    $key = $as_shortnames ? relationship_get_shortname( $s[$columna] ) : $s[$columna] ;
    $a[$key] = $s[$columnb];
  }
  return $a;
}

/**
 * Check the given statement is ready to be stored.
 * 
 * As well as validating the values, this actually converts statements made
 * using fully qualified names to our local shortname identifiers.
 *
 * This may be a little short-sighted, but it makes editing a lot
 * easier, and can be turned off here when we want to go to
 * full namespaces.
 * Transforms both predicate and object (but not subject) into canonical
 * shortnames from URLs (if the identifier can be found in the DB)
 *
 * @param a statement, passed and modified by ref
 * @return false if statement is invalid in some way
 */
function relationship_validate_statement(&$s) {
  if(! $s['subject'] || ! $s['predicate'] || ! $s['object'] )
  { 
    return false; 
  }
  if(gettype($s['subject'])!='string'){return false;}
  if(gettype($s['predicate'])!='string'){return false;}
  if(gettype($s['object'])!='string'){return false;}

  if(! variable_get('relationship_long_names', FALSE) ){
    if(is_uri($s['predicate'])){
      $s['predicate'] = relationship_get_shortname($s['predicate']);
    }
    if(is_uri($s['object'])){
      $s['object']  = relationship_get_shortname($s['object']);
    }
  }
  
  return $s;
}

/**
 * Save a given statement triple in the datastore.
 *
 * @param $statement 
 *        A statement is an assoc array defining 'subject',  'predicate',  'object',
 *        and optionally 'title' and 'uid' @see relationship_statement()
 * @see relationship_statement()
 * @see relationship_query()
 * @todo dupe checking
 * @todo datastore interface
 */
function relationship_save_statement($statement) {
  if(! relationship_validate_statement($statement)){
    debug("Not saving invalid statement: ".print_r($statement,1) ,3); 
    return FALSE;
  }
  $ds = _relationship_datastore();
  return $ds->add($statement);
}

/**
 * Basically a shorthand way to create and save a statement in one go.
 * @see relationship_statement()
 * @see relationship_save_statement()
 */
function relationship_set($subject,  $predicate,  $object,  $title=NULL,  $uid=NULL) {
  return relationship_save_statement(
    relationship_statement($subject,  $predicate,  $object,  $title,  $uid)
  );
}


/**
 * Attempt to make a SPARQL-type query on the datastore.
 *
 * relationship_query() does not automatically get titles.
 * If you wish to link to the results of a query, call relationship_get_titles()
 * on the result yourself before proceeding.
 *
 * @param $subject optional
 * @param $predicate optional
 * @param $object optional
 * @param $column optional
 * @returns a collection of statements. Even if empty, the return value is an
 * array.
 */
function relationship_query($subject=NULL, $predicate=NULL, $object=NULL, $column=NULL) {

  // Allow the actual subject resource object to be passed in by reference
  // (is this used much?)
  if (is_object($subject) ) {
    if ($subject->uri) {
      $subject=$subject->uri;
    } else if ($subject->nid) {
      $subject=$subject->nid;
    }
  }

  debug('Querying <strong>'._relationship_hash_key(relationship_statement($subject,$predicate, $object)).'</strong>' , 2);

  // Informative paranoia is vital in an API
  // Do some assertions

  // Initialize, and check our datastore
  $ds = _relationship_datastore();

  if(! $subject_statements = $ds->query($subject,$predicate, $object)){
    $subject_statements = array();
  }

  // alias (sameAs) support?
 
 
  // If the datastore doesn't do it's own inverse support, I'll do it myself
  // note, I don't cache or anything
  if(! $ds->supported_characteristics(PREDICATE_INVERSE_OF) ){
    debug("Doing reverse lookup to get more data about $subject $predicate $object",2);
    $query = relationship_statement($subject, $predicate, $object);
    $inverse_statements = relationship_query_inverseOf($query);
    if($inverse_statements){
      $subject_statements = array_merge($inverse_statements,$subject_statements);
    }
  } else {
    debug("NOT Doing reverse lookup to get more data about $subject . Datastore should have done it.",2);
  }

  ////////////////////////
  // See if there is a custom hook for executing this query supplied by a
  // module or library.
  // If so, it can do all the work. 
  // Code will take precedence over any database lookups.
  // todo define this hook interface more
  // - should always return an array, even if empty
  ////////////////////////
  $custom_lookup = 'relationship_lookup_'.$predicate;
  if ( $predicate && function_exists($custom_lookup) ) {
    // A pluggable ontology has defined its own 
    // lookup routine,  let it take over from here.

    debug("Calling custom lookup function $custom_lookup($subject) ", 1);
    $custom_statements = $custom_lookup($subject);
    debug("Custom lookup $custom_lookup() returned  ".print_r($custom_statements, 1), 2);

    $subject_statements = array_merge($custom_statements,$subject_statements);

    // That'll do for now. Cache results in the current context and return!
    
    // Not saving deduced values in the database permanently... yet
    // But if we did, it would be:
    // $ds->add_array($subject_statements, $custom_lookup);
  }
  debug('Query <strong>'._relationship_hash_key(relationship_statement($subject,$predicate, $object))."</strong> returned ".($subject_statements? "<ul><li>".join("</li>\n<li>",array_keys($subject_statements))."</li></ul>" : "no results")  , 2);

  // more paranoia
  // I can't imagine how this could happen, but with pluggable datastores and lookups, anything is possible
  foreach($subject_statements as $key => $statement){
    if(! relationship_validate_statement($statement)){
      trigger_error("
        Invalid statement returned from datastore or query. 
        Something is wrong in the database or code.: <br/>
        Query was "._relationship_hash_key(relationship_statement($subject,$predicate, $object)).
        "Result was: <pre>".print_r($statement,1)."</pre>");
        debug_pre($subject_statements);
        unset($subject_statements[$key]);
    }
  }


  // Filter to just the results
  if($column && $subject_statements){return relationship_get_column($subject_statements,$column);}

  return ($subject_statements) ? $subject_statements : array();
}


function relationship_delete_statement($statement){
  $ds = _relationship_datastore();
  return $ds->delete($statement);
}

/**
 * Remove associations of a resource (node) to its terms.
 *
 * Used when updating node statements to start afresh.
 *
 * Optionally, also delete all references to it - this should only
 * be done when deleting an item and forgetting it ever existed.
 *
 * $param $subject nid, path or URI
 * $param $extreme_prejudice Wipe out anyone who ever came into contact with this ...
 */
function relationship_resource_delete_statements($subject, $extreme_prejudice=FALSE) {
  debug("Deleting all links and statements about '$subject'",2);
  if(! $subject){
    debug("Deleting everything about nothing. probably an error somewhere");
    return;
  }
  $ds = _relationship_datastore();
  $ds->delete($subject);  
  
  if($extreme_prejudice){
    if($rows_changed = $ds->delete(NULL,NULL,$subject) ){
      debug(t("$rows_changed statements REFERRING TO $subject also deleted."),2);  
    };
  }
}





/**
 * Return a handle on the datastore
 */
function _relationship_datastore() {
  static $ds;  
  if (!$ds) {$ds = new relationship_datastore();}
  return $ds;
}


/**
 * Generate a hash key to index the statements on.
 * 
 * Used to make sure we neither double up or overwrite stuff we already know
 * when recursing. Also useful shorthand when data-dumping, 
 * it's a string representation of the statement, 
 * with some reference to N3, but without problematic <>
 * Used mainly by the datastore, but also a little as other functions build up
 * their own statement collections.
 * @param $statement a relationship_statement() 
 * @param $level the number of elements long the statement should be. An optimizing
 *               datastore may index partial lookups using a partial hash
 */
function _relationship_hash_key($statement,$level=3) {
  if(!$statement){return NULL;}
  
  $statement['object'] = substr(preg_replace("|[^A-Za-z0-9]+|","_",$statement['object']),0,50);
  
  $h = ($statement['subject'] ? $statement['subject'] : '?' ).STATEMENT_SEPARATOR;
  $h .= (($level>1 && $statement['predicate']) ? $statement['predicate'] : '?').STATEMENT_SEPARATOR;
  $h .= (($level>2 && $statement['object']   ) ? $statement['object']    : '?');
  return $h;
}



/**
 * Lookup statements that imply the given query via the inverse rule.
 *
 * Predicate may be blank, which would mean anything that says anything about the subject.
 */
function relationship_query_inverseOf($query){
  if(!$query){debug('no query? this is a waste of time',2);return;}
  
  $statements = array();
  $query = relationship_inverse_statement($query);

  if($query){
    $ds = _relationship_datastore();
    $inverse_statements = $ds->query($query);
    if($inverse_statements){
      debug('Found inverse statements: '.join(', ',array_keys($inverse_statements)),3);
      foreach($inverse_statements as $i => $statement){
        if($s = relationship_inverse_statement($statement)){
          //
          // oddness. What if we find more than one 'inverse'?
          // it's causing pain, but what happens if I support it?
          // what will the logical implications be?
          if(gettype($statement['predicate']) == 'array'){
            foreach($statement['predicate'] as $i => $p){
              $alt_s = $statement; $alt_s['predicate']=$p;
              $statements[_relationship_hash_key($alt_s)] = $alt_s;
            }
          }
          else
          $statements[_relationship_hash_key($s)] = $s;
        }
      }
    } else {
      debug("No results from inverse query "._relationship_hash_key( $query ),2); 
    }
  }
  // else {
  //  debug("There is no inverse to the query. ". _relationship_hash_key( $query ) ,3);
  // }
  return $statements;
}


/**
 * Reverse the operands and inverse the predicate for a statement.
 *
 * Eg turn A IsChildOf B into B HasChild A If you already know the reciprocal,
 * pass it in, otherwise it will be retrieved.
 *
 * Helper for relationship_query_inverseOf()
 *
 * Inverse of a statement with a blank predicate returns as an inverse
 * statement with a blank predicate. This is (almost) meaningless as a 
 * statement, but useful as a query.
 * The inverse of the question
 * {this, ?, ?}  - tell me everything about this
 * is
 * {?, ?, this}  - tell me everything that MENTIONS this
 *
 * @param $statement @see statements
 * @param $reciprocal predicate term definition @see predicates
 * @return modified statement,  also changes it by reference. 
 *         Returns null if no inverse predicate exists
 */
function relationship_inverse_statement(&$statement, $reciprocal='') {
  if (! $reciprocal && $statement['predicate']) {
    $term_def = relationship_get_term($statement['predicate']);
    if(!$term_def){ return; } // no such relationship
    $reciprocal = $term_def->inverseOf;
  }
  if(gettype($reciprocal) == 'array'){
    trigger_error("Reciprocal of ".$statement['predicate']." is apparently ".print_r($reciprocal,1)." - this cannot be, there can be only one inverse. This should probably be repaired.");
  }  
  if ($reciprocal || (!$statement['predicate']) ) {
    $recip_statement = array(
      'subject'   => $statement['object'],
      'predicate' => $reciprocal,
      'object'    => $statement['subject'],  
      'inverseOf' => $statement['predicate'],  
      'title'     => '',
      'depth'     => $statement['depth'] +1,
      'why'       => 'inverse of '.($statement['predicate'])?$statement['predicate']:'wildcard lookup'
   );
  }
  return $statement = $recip_statement;
}



/**
 * Return all available predicate terms, or a selected one of them.
 *
 * Indexed by term name,  as an array of term definition OBJECTS with the 
 * properties (name,  label,  definition, weight) defined. 
 *
 * Internally cached,  so call it as often as you like.
 *
 * Predicates are resources whose type is CLASS_PROPERTY
 *
 *
 * @param $name Term identifier (currently terms are keyed on shortname - this may change)
 * @param $and_statements - a flag whether to retain the terms meta statements internally
 *                          in a ->statements array. Usually unwanted, but useful when
 *                          analyzing a term directly. Using this also bypasses caching;
 * @return array of term objects,  or just one term object if requested by name.
 * @see predicates
 * @ingroup api
 */
function relationship_get_terms($name='', $and_statements=FALSE) {
  static $terms;
  if(! $terms){ $terms=array(); }

  // also cache the failed lookups. This will save us from looking again, but will NOT
  // return the auto-instantiated definitions in the main list.
  static $null_terms;

  static $recursion_panic; // Don't want to get carried away now ...
  if($recursion_panic){return $terms ? ($name ? $terms[$name] : $terms) : relationship_term_stub($name);}
  $recursion_panic = TRUE; // If we see this again in this lifetime, it'll be too soon


  // Some SQL magic could probably clean this up a lot
  // The terms cache is flushed every time $and_statements is fetched.
  // This is a little TOO frequent, but only happens when the admin is actually
  // editing predicates terms, so it's appropriate enough.
  
  if ( $and_statements || ! $terms ) {

    if( (! $and_statements) && ($cached = cache_get(RELATIONSHIP_TERMS_CACHE_KEY)) ){
      // fetch from cache unless we need the full story.
      $terms = unserialize($cached->data); 
    } else {
      debug('Fetching predicates afresh, as '. ($and_statements?' we need to load all because the statements were requested, and we dont usually cache all them':' cache was empty'),2);
      if($and_statements){debug("Fetching all statements on all terms is a bad idea - Avoid it as much as possible",3);}

      // Try to find all 'predicates' - things of type 'Property'
      $predicates = relationship_get_resources_matching_query(PREDICATE_TYPE, CLASS_PROPERTY, $and_statements);

      foreach($predicates as $term){
        // Validate and fill in any gaps we may assume a 'term' should have
        if(!$term->uri){
          // Need to remember what we are talking about
          $term->uri = $name;
        }

        if(! is_string($term->identifier)){
          // RDF imports don't explicitly inform us what the shortname is - I'll guess
          $term->identifier = relationship_guess_shortname($term->uri);
        }
        
        if(! is_string($term->name)){
          // TODO - sort out whether I use 'name' or 'identifier''. I forget.
          $term->name = $term->identifier;
        }

        if(! $term->label){
          watchdog('relationship', t("
            Invalid predicate property found when retrieving terms 
            from the database ( relationship_get_terms() ). 
            Apparently, \"'.$term->uri.'\" is a RDF Property, 
            but it doesnt define the required \"label\" parameter. 
            It should be removed or fixed immediately. 
            This may be caused by some non-property resource claiming 
            it is of \"type\" \"property\""));
          debug_pre($term,1);
        } else {
          $terms[$term->identifier] = $term;
          debug("Retrieved (uncached) Definition of term ".$term->uri." - id:'".$term->identifier."'", 1);
        }
      }

      debug("Updating the 'predicate terms' cache. Terms were not statically available, and_statements was '$and_statements' and cached was '$cached' ",1);
      cache_set(RELATIONSHIP_TERMS_CACHE_KEY,serialize($terms)) ;
    } // not cached (or needed statements)
  } // not static (or needed statements)

  if($name){
    debug("Fetching the short name of term with the id '$name' ",3);
    $name = relationship_guess_shortname($name);
  } else {
    debug('Returning all terms',3);
  }
  
  // Asked for something that doesn't look like a term ... but may be
  // Unsure about my term selection process above
  // NOTE, this routine may throw some red herrings during the bootstrap process
  //
  if($name && (! isset( $terms[$name] )) && (! isset( $null_terms[$name] )) ){
    debug("'$name' is being requested as if it's a predicate, yet is not in the database as a recognised term. Using it as a property anyway, but our ontology is probably incomplete. $and_statements",1);
    $null_terms[$name] = (object)array(); // setting it here will preclude a dangerous long-winded recursive loop
    
    $term = relationship_get_resource_by_name($name, relationship_get_property_map(CLASS_PROPERTY), $and_statements ) ;
    if($term){
      debug("'$name' is not a term, However I do have SOME information matching that description, using that.",2);
    } else {
      debug("I have no idea what '$name' is supposed to mean. Using it as a placeholder predicate term for now.",2);
    }
    // this is probably an inconsistancy, so should I cache it here?
    /* */
    $terms[$name] = $term; // cache even if blank - avoid looking again.
    //  cache_set($terms_cache_key,serialize($terms)) ;
    /* */
  }
  # debug_pre(array_keys($terms),3);
  
  $recursion_panic = FALSE;
  return $name ? ($terms[$name] ? $terms[$name] : $null_terms[$name] ) : $terms;
}

/**
 * Empties the terms cache.
 *
 * Required when it seems a change way made that might change our ontology.
 * @TODO Move to administer_predicates? It's only called when predicates
 * change, which should be seldom.
 */
function relationship_flush_terms() {
  // clear relationship_get_terms() 
  cache_clear_all(RELATIONSHIP_TERMS_CACHE_KEY);
  debug("Flushed the internal 'terms' cache",2);

  // This refreshes the cached list of properties of properties
  relationship_get_all_characteristics(CLASS_PROPERTY,TRUE);

  // This refreshes the cached drupal->canonic property label maps, 
  cache_clear_all(RELATIONSHIP_PROPERTY_MAP_CACHE_KEY);
  relationship_get_property_map(NULL,TRUE);
  
  // Discard our shortname cache
  variable_set('relationship_shortnames',NULL);
}


/**
 * Return the term object matching a term identifier (usually its shortname).
 *
 * Synonym for relationship_get_terms() used in single lookup mode.
 * 
 * If a requested term definition does not exist, a dummy placeholder
 * representing what you asked for is returned. This is to support
 * orphaned relationships which still exist as statements even if the
 * supporting module has gone away.
 *
 * @param $name string shorttname of the term
 * @param $and_statements bool whether to retain all the statement declarations that define this predicate.
 * @returns Object defining a terms properties (label and characteristics)
 * @see predicates
 * @see relationship_get_terms()
 * @see relationship_get_resource_by_name() - which is identical but without
 * caching
 */
function relationship_get_term($name,  $and_statements=FALSE) {

  if( $and_statements ) $term = relationship_get_resource_by_name($name, NULL, $and_statements );
  else $term = relationship_get_terms($name);

  if (! $term) { 
    // Allow terms referring to disabled ontologies to still show up - in a restricted manner.
    // This is probably unwanted in practice,  put it to support displaying all metadata when
    // the database is in an unstable (development) state, or to create placeholders for
    // unknown quantities during the first bootstrap from first principles.

    debug("'$name' is not (yet) in the predicate table as a recognised term, This is annoying,  but not fatal. Returning a made-up definition",2);
    $term->name = $term->label = $term->uri = $name; $term->type = CLASS_PROPERTY;
  }
  return $term;
}


function relationship_term_stub($name){
  $term = new stdClass();
  $term->name = $term->label = $term->uri = $name; 
  $term->type = CLASS_PROPERTY;
  return $term;
}




/**
 * Retrieve all statements about a given identfier
 * as a structured array.
 *
 * If multiple statements have the same predicate,
 * they will be gathered as an array.
 * (Be careful if you assume there can be only one value, 
 * this enacts very loose typing, and ordinality is not enforced.)
 *
 * If a mapping template is given, certain predicates 
 * will be renamed into keys on the resulting assoc array.
 * If no mapping template is given, ALL statements will become properties
 *
 * @parameter $name shortname. Or uri.
 * @return an object structure representing the information about the named
 * resource. Null if totally unknown.
 */
function relationship_get_resource_by_name($name, $property_map=array(), $and_statements=FALSE) {
  $id = relationship_get_longname($name);
  debug("fetching resource '$name' [$id]",2);
  if($id){
    $statements = relationship_query($id);
  } else {
    debug("No id found for the resource identified as $name. This is a bit odd, probably means incomplete data. Returning just the name as an ID",3);
    $statements = relationship_query($name);
  }

  if(! variable_get('relationship_long_names', FALSE)){
    // Also query on shortname, just to be complete
    $shortname = relationship_get_shortname($id); 
    if($shortname != $id){
      debug("Also looking up data about the shortname $shortname",2);
      $statements = array_merge( relationship_query($shortname) , $statements);
    }
  }
  if(!$statements) return NULL;

  if(!$property_map){
    // See if we know what type it is,
    // If it's recognised, use that mapping
    $property_map = relationship_get_property_map(relationship_get_property(($id?$id:$name),PREDICATE_TYPE));
  }

  return (object) relationship_statements2array($statements, $property_map, $and_statements);
}


/**
 * It's quite common we want to retrieve all data about all objects
 * meeting a cetain condition
 *
 * This wraps up two major steps -
 * What resource id matches what criteria (eg everything with type=ontology)
 * Then retrieves the definition of each resource.
 *
 * This helper function is not flash at the moment, but is out on its own 
 * because it's ripe for SQL optimization - all it needs is a join
 */
function relationship_get_resources_matching_query($predicate, $object=NULL, $and_statements=FALSE) {
  $resource_ids = relationship_query(NULL, $predicate, $object,'subject' );
  $resources = array();
  foreach($resource_ids as $id){
    // load everything there is to know about this predicate
    $resources[$id] = relationship_get_resource_by_name($id, NULL, $and_statements );
  }
  return $resources;
}


/**
 * Utility shorthand functions, quick versions of the major actions
 */

/** 
 * Returns the object of a subject,predicate query, 
 * the value of a relationship on a resource.
 *
 * A convenience simplification of relationship_get_statement()
 * Just returns one string result, not the full statement.
 * Null if no value found.
 */
function relationship_get_property($subject, $predicate) {
  $s = relationship_get_statement($subject, $predicate, FALSE);  
  return $s['object'];
}

/** 
 * Return all given values for a lookup of a relationship
 * on a resource.
 *
 * A convenience simplification of relationship_get_statement()
 * Retrieves all matches, but only returns the object value, not the full statement.
 * Empty array if no value found.
 */
function relationship_get_properties($subject, $predicate) {
  return relationship_query($subject, $predicate, NULL, 'object');  
}

/**
 * Return one statement matching a given RDF Query.
 *
 * Same as relationship_query(),  but return just one statement 
 * of the given type for the given subject. 
 * A convenience method to save popping the array.
 * @see relationship_query()
 *
 * If there are priority issues within the predicate - what should be the primary
 * statement returned in this case - they must be resolved by the ontology itself.
 * This function just returns the first thing it finds. That is to say,  the 
 * one shallowest in the recursion 'depth'.
 *
 * TODO - this is not optimised,  it retrieves all and returns one,  should really
 *        return as soon as one is found...
 *
 * @param $subject usually a node id,  can also be a path (eg taxonomy/term/5) or even a URI
 * @param $predicate bool return only this type of statement
 * @param $max_recursion_level int When looking up relationship,  follow reciprocal and recursion rules this far,  the returned list of statements will include implied statements also
 * @returns a 'statement' structures @see statements
 */
function relationship_get_statement($subject,  $predicate, $and_title=TRUE) {
  $statements = relationship_query($subject,  $predicate);
  if (is_array($statements) ) {
    // return the shallowest successful result
    uasort($statements,  'relationship_sort_statements_by_depth');
    $statement = array_shift($statements);

    if($and_title){
      relationship_get_object_title($statement);
    }
    
    return $statement;
  }
}

/* Util for sorting */
function relationship_sort_statements_by_depth($a, $b) {
  return ($a['depth'] < $b['depth']) ? -1 : 1;
}


/**
 * Return the local shortname of a term associated with a given URI.
 *
 * Imported data is likely to come with fully-qualified predicate identifiers,
 * (longname)
 * See if we know about the term it's talking about (shortname).
 * Scan the local available term definitions to see If we can find a match.
 * If not,  return the full version again.
 * 
 * Currently cached in drupal variables. There should only be a dozen or two.
 *
 * @see relationship_guess_shortname() which just hashes it from the URI
 * @see get_longname()
 */
function relationship_get_shortname($uri='') {
  if(! ($uri && is_string($uri)) ){debug('relationship_get_shortname() called with null or invalid argument, this is almost certainly an error somewhere earlier.');return '';}  

  static $shortnames;
  if(!$shortnames){
    $shortnames = variable_get('relationship_shortnames',array());
  } 
  if(isset($shortnames[$uri])){return $shortnames[$uri];}


  if(in_array($uri,$shortnames)){
    debug("$uri is already a shortname, you idiot",0);
    $shortnames[$uri] = $uri;
    variable_set('relationship_shortnames',$shortnames);
    return $uri; 
  }
  // Haven't found it yet. Look up, then cache.

  // See if it's in the 'local' namespace
  $shortname = relationship_get_localname($uri);
  
  if(! $shortname){
    $shortname = relationship_get_property($uri, PREDICATE_NAME);
  }
  
  if($shortname){
    $shortnames[$uri] = $shortname;
  }
  else
  { 
    // should I cache failures?
    // This may lead to unwanted false names
    $shortnames[$uri] = relationship_guess_shortname($uri);
    debug("Warning. '$uri' has no defined shortname in the DB. Deducing (and remembering) that '".$shortnames[$uri]."' is good name for it, but this may be wrong or lead to confusion as possibly merges all concepts called ".$shortnames[$uri],1);
  }
  variable_set('relationship_shortnames',$shortnames);
  return $shortnames[$uri] ;
}

/**
 *  Given a shortname, retrieve the associated longname or uri.
 *
 * If not,  return the given short name.
 * Also works over arrays of names, so it could make a shorthand statement cannonical in one go.
 * If it's already a URI, return that.
 * Otherwise, lookup to see if this is being used as an identifier for anything.
 * Otherwise, if the name appears to be a key referring to a locally defined
 * resource, the longname is the local schema URI + the identifier.
 * @todo see if it's appropriate to use the cached shortname array as a rev
 * lookup.
 * @see get_shortname()
 */
function relationship_get_longname($name='') {
  if (is_array($name)){
    //  Overload this function. Loop over all given 'names'.
    return array_map( 'relationship_get_longname', $name);
  }

  if(! is_string($name)){
    debug('invalid argument ['.gettype($name).'] to relationship_get_longname(). Expected a string. '.print_r($name) ,1);
    return NULL;
  };
  
  static $longnames;
  if ($id = $longnames[$name]){
    return $id;
  }

  if(is_uri($name)){return $name;}

  // not cached, look it up.
  $names = relationship_query(NULL, PREDICATE_NAME, $name, 'subject');
  if($names){ 
    $longnames[$name] = array_shift($names);
    if(count($names)>1){debug("Odd, There appears to be more than one resource with the same name '$name' : ".print_r($names,1));}
  } else {
      $longnames[$name] = $name;
  }
/*
  if($longnames[$name] == $name){
    // So far we haven't learnt anything.
    // See if we know any more about this name as a subject resource
    $owner_ontologies = relationship_get_properties($name,PREDICATE_IS_PART_OF);
    // if this appears to be a mention of an internal predicate, prepend the local schema space to it
    if($owner_ontologies){
      debug("$name appears to be a locally defined resource (it's listed as being a member of an ontology we know of, although it didn't seem to have a valid URI). Resolving its path to refer to this server.",2);
      $longnames[$name] = url(variable_get('ontology_rdf_schema_path', 'relationship/schema').variable_get('ontology_rdf_namespace_modifier', '/'),NULL,NULL,TRUE).$name;
    } 
  }
*/

  return $longnames[$name] ;
}

/**
 * Return the shortname of any anonymous identifier.
 * Like get_shortname, but only returns a result if the URI appears to be in the
 * default namespace - it begins with the locally defined schema path.
 * @returns the local identifier. NULL if it's not local.
 */
function relationship_get_localname($uri){
  static $local_schema_path;
  if(! $local_schema_path){$local_schema_path = url( variable_get('relationship_rdf_schema_path', 'relationship/schema') . variable_get('relationship_rdf_namespace_modifier', '/'),NULL,NULL,TRUE);}

  if(strstr($uri,$local_schema_path)){
    // Looks like it should be declared as a local name 
    debug("Should truncate $uri, it contains $local_schema_path",3);
    $shortname = substr($uri,strlen($local_schema_path));
    return $shortname;
  } else {
    return NULL;
  }
}

/**
 * Return whatever we can find as a readable (shortish) name for a resource.
 * If a resource has any of the properties:
 * - label
 * - name
 * - identifier
 * - title
 * - uri
 * Return the first one we find.
 */
function relationship_get_display_label($resource){
   return $resource->label? $resource->label :
    ($resource->name? $resource->name : 
      ($resource->identifier? $resource->identifier :
        ($resource->title? $resource->title :
          $resource->uri
        )
      )
    ); 
}

/**
 * Given a long namespace-qualified term URI, return what looks like
 * the actual term. By way of string-munging, not lookups.
 *
 * Funny the term definitions themselves don't define a shortname!
 *
 * Eg, from 
 * relationship_guess_shortname('http://purl.org/dc/elements/1.1/title')
 * Return 'title'
 *
 * Internally, I use shortnames as the predicates.
 * Although they COULD have
 * been URIs all the way it was really had to read.
 * This is different from relationship_get_shortname() because it
 * does not 'get' it from a lookup, it invents it on the fly.
 *
 * @param $uri longname
 * @return string shortname
 * @see relationship_get_shortname()
 * @see relationship_get_longname()
 * @ingroup relationship_api
 */
function relationship_guess_shortname($uri) {
  // Have to chop the URI to deduce it.
  // truncate either at the trailing Hash or the last slash - this may cause trouble.
  if ( preg_match('|[#/]?([^/#]+)$|',$uri,$uri_bits) ){
    return $uri_bits[1];
  } else {
    debug('Failed to deduce a shortname from the property definition "'.$uri.'". This could be a problem. Using full the URI (is it a URI?) as a predicate instead',2);
    return $uri;
  }
}



/**
 * Convert an assoc array to a collection of statements
 * about something.
 *
 * From a php properties array (very much like turtle syntax) :
 *
 *   $thisItem = array(
 *        'name' => 'joe',
 *        'age' => '55',
 *   );
 *
 * We learn that:
 *
 * $statements = array(
 *  [] = array(
 *      'subject' => 'thisItem',
 *      'predicate' => 'name',
 *      'object'    => 'joe',
 *    ),
 *  [] = array(
 *      'subject' => 'thisItem',
 *      'predicate' => 'age',
 *      'object'    => '55',
 *    ),
 *  );
 *
 * @see <http://www.dajobe.org/2004/01/turtle/>
 * @see relationship_statements2array()
 * @todo rename this turtle2triples() or something?
 * 
 * @param  $properties assoc array
 * @param $subject The 'subject' identifier that these statements are being made about. A
 *                 string, not the object. 
 *                 If not defined, either the 'uri' or the 'name' 
 *                 provided in the data will be used instead. (in that order)
 * @param $import_template An assoc array matching keys used in the array to
 *                 predicate  terms - if any translation is needed.
 * @param $strict  ONLY analyse statements listed in the import template.
 *                 This is handy when abstracting a limited amount of info out
 *                 of a complex object (like a node). 
 *                 If off, ALL properties will be converted to triples
 * @return unindexed array of statements
 */
function relationship_array2statements($properties, $subject=NULL , $import_template = array() , $strict = FALSE , $why = 'Internal Object Conversion') {
  $statements = array();

  if(! $subject){
    $subject = $properties['uri'] ? $properties['uri'] : $properties[PREDICATE_NAME] ? $properties[PREDICATE_NAME] : $properties['name'];
  }
  if(! $subject){
    debug("An anonymous subject, what am I supposed to say about it?",2);
    return $statements;
  }
  
  if(! $import_template){
    $import_template=relationship_get_property_map($properties['type']);
  }
  
  $key_mapping = array_flip($import_template);

 #debug('Importing data array <pre>'.print_r($properties,1).'</pre> into statements using renaming map <pre>'.print_r($key_mapping,1).'</pre>' ,0);
  
  foreach($properties as $predicate => $object){
    if(isset($key_mapping[$predicate])){
      // rename certain array keys to our custom terms
      debug("Mapping the resource->$predicate property ($object) into a statement about '".$key_mapping[$predicate]."'" ,2);
      $predicate = $key_mapping[$predicate];
    } else if ($strict){
      // and don't randomly add stuff if we are 'strict'
      continue;
    }
    
    if( (!$predicate) || ($predicate =='statements') ){
      // avoid the 'statements' array that might already be embedded in this object,
      continue;
    }
      
    
    if($predicate){
      if($object && is_string($object)){
        // tidy the $object string (may have whitespace from the code layout)
        $object = trim(preg_replace('/ +/',' ',$object));
        $s = relationship_statement($subject,  $predicate,  $object, NULL, 0, $why);
        $statements[_relationship_hash_key($s)] = $s;
      } 
      else if(is_array($object)){
        // multiple values to one predicate are also legal shorthand
        // 'Robert' = array( 'sameAs' => array('Bob','Bobby') ) ;
        // gets broken into two statements.
        foreach($object as $ob){
          $s = relationship_statement($subject,  $predicate,  $ob, NULL, 0, $why);
          $statements[_relationship_hash_key($s)] = $s;
        }
      }

    }
  }
  return $statements;
}



/**
 * Given a load of statements, build a picture about the subject.
 *
 * Cast all the given relationships and data as array properties.
 * Collates a 'thing' from a loose collection of assertions.
 * Used for quick n dirty instantiation of a data cloud to something I can use.
 *
 * @param $statements a collection of statements, all about one subject.
 * @param $property_map an array indexing any renaming of predicates to properties that may be required.
 * @return an array of structured, indexed information
 * @see relationship_array2statements()
 * @ingroup relationship_api
 */
function relationship_statements2array($statements, $property_map=array(), $and_statements=FALSE) {
  if(! is_array($statements)){return array();}
  
  $thing = array();
  
  debug('Translating from statements <pre>{  '.join(",\n  ",array_keys($statements)) .'}</pre> to a structured array using name mapping rules : <pre>'.print_r($property_map,1).'</pre>',2);
  
  foreach ($statements as $i => $statement){
    if(! is_array($statement)){ debug("Invalid statement (non-array) in statements collection"); continue; }
    $p = $statement['predicate'];
    if(! $p){ debug("Invalid statement (no predicate) in statements collection",1); continue; }

    if(isset($property_map[$p])){
      // change its name according to the rules.
      
      if(is_string($property_map[$p])){
        // set it as a string - but what if it's already set?
        
        $thing[$property_map[$p]] = $statement['object'];
      }
      else if(is_array($property_map[$p])){
        // it's a collection of similar statements
        $thing[$property_map[$p]][] = $statement['object'];
      }
      else {
        debug("Assigning a non-string property? Trying to set {$property_map[$p]} to {$statement['object']} could be trouble");
        $thing[$property_map[$p]] = $statement['object'];
      }

    } else {
      // this property is not in the naming map

      if(! $thing[$p] ){
        $thing[$p] = $statement['object'];
      } else {
        // auto-extension of properties into an array when a dupe is found.
        // This may casuse confusion down the track. Just check your strings
        // before working with them, they may be arrays!

        if (is_array($thing[$p])){
          $thing[$p][] = $statement['object'];
        } else {
          $old = $thing[$p];
          $thing[$p] = array($old,$statement['object']);
        }
        
      }
    } // not special

    if($and_statements){
      // retain the explicit statements in the object - useful if editing, but usually unwanted
      $thing['statements'][$i] = $statement;
    }
    
  } // each

  if($statement['subject']){
    $thing['uri'] = $statement['subject'];
  }
  debug_pre($thing,2);

  return $thing;
}







/**
 * Return the property map for a Drupal-defined content type.
 *
 * Although I can represent core Drupal data structures (node, user etc)
 * using formal terminology, sometimes the properties need to be renamed.
 * This function returns a mapping array for various existing data types.
 *
 * Properties NOT in this list will be dealt with as normal arbitrary listed
 * predicates, which get serialized as-is, but don't get any special 
 * treatment.
 * Properties that ARE listed here are expected 
 * to be serialized individually by the Drupal form interface.
 * - I won't save them.
 *
 * The arrays are also used as for mapping Drupal-specific terminology to 
 * other terms from existing namespaces
 * otherns:term => drupal:term 
 * predicate => property
 *
 * eg, 'name' isn't a real RDF term, we'll save it
 * as 'identifier' but let Drupal keep calling it the 'name' property.
 *
 * Use this function call to see if a statement has a similar Drupal property
 * that should get filled in.
 * (Reading from triple-store into node def)
 *
 * Use array_flip to find the inverse, and make sure Drupal
 * 'parameters' get saved out of the object as 'statements'.
 * (When serializing a node into the triple store)
 * Note that this may cause some data compression when converting to cannonical
 * format. Multiple Drupal labels can be associated with the same DC label,
 * ($node->flexinode-3, $node->myname, $node->title) may all collapse 
 * to equal DC:title but not vice versa.
 *
 * This method can be used to 'cast' an existing node type (with arbitrary
 * fields) into a set of statements. It supports arrays also, so if you map
 *
 *      'dc:subject'  => 'taxonomy'
 *
 * then 
 *      $node->taxonomy = ['sex','drugs','rock&roll']
 *
 * will be mapped to the corresponding
 *
 *      node : dc:subject : 'sex'
 *      node : dc:subject : 'drugs'
 *      node : dc:subject : 'rock&roll'
 *
 * statements and serialized.
 *
 * The actual mapping is set by making the statement 
 *      'dc:subject'  'drupal:equiv' 'taxonomy'
 * in the datastore.
 * 
 * @todo allow custom modules to populate this, they'll need it. 
 * 
 * The $drupal_type parameter is now deprecated, the DRUPAL_EQUIV 
 * characteristic now configures this.
 * This is aggresively cached, so if changes are made that may affect it
 * (like a term being assigned a DRUPAL_EQUIV value) this must be flushed.
 *
 * @see relationship_get_resource_by_name()
 * @see relationship_statements2array()
 * @see relationship_array2statements()
 * @param $flush if set, the cache is flushed and a new array generated
 */
function relationship_get_property_map($drupal_type='') {
  static $maps;
  if(isset($maps[$drupal_type])){
    debug('returning (cached) property map '.print_r($maps[$drupal_type],1),3);
    return $maps[$drupal_type];
  }
  $cached = cache_get(RELATIONSHIP_PROPERTY_MAP_CACHE_KEY);
  if($cached->data){
    $maps = unserialize($cached->data);
    debug('returning cached property map '.print_r($maps,1),3);  
  }

  if(! $maps[$drupal_type])
  {
    debug("Calculating new property map for objects of type '$drupal_type'",1);  

    static $loop;
    // Due to the self-referential nature of this, we may get recursion panic
    // on bootstrapping. If so, just return the hardcoded basics.
    // 'name' is dc:identifier and 'description' is rdf:comment 

    $built_in_mappings = array(PREDICATE_NAME => 'name', PREDICATE_COMMENT => 'description' );
    if($loop){
      debug("Recursion panic when retrieving property map, returning hardcoded property map",1);
      return $built_in_mappings;
    } // recursion panic
    $loop = TRUE;
    
    $statements = relationship_query(NULL,PREDICATE_DRUPAL_EQUIV,NULL,NULL);

    // other_ns:term => drupal:property
    $map = relationship_get_assoc_column($statements,'subject','object',TRUE);
    $map = array_merge($built_in_mappings,$map);

    // Special case. The 'name' of most resource items is its identifier,
    // but not Drupal nodes... a node 'name' is its creator :-[
    if($drupal_type == CLASS_NODE){
      // make this a subclass query
//      $map['creator'] = 'name';
    }

    // uri is not a useful value to put into the database, 
    // the uri of a resource is implicit in the resource def
    $map[''] = 'uri'; 

    $maps[$drupal_type] = $map;
    cache_set(RELATIONSHIP_PROPERTY_MAP_CACHE_KEY,serialize($maps)) ;
  }
// dsm($maps);
  $loop = FALSE;
  return $maps[$drupal_type];
}



/**
 * Convert the information contained in the given resource to relationship statements, and save them.
 *
 * An 'import table' defining how object properties map to statement predicates will be retrieved and used if possible.
 * @see relationship_get_property_map()
 */
function relationship_save_resource($edit, $subject = '', $import_template = array(), $why='imported from save_resource()' ){
  $edit = (array)$edit;

  if(!$import_template){
    $import_template = relationship_get_property_map($edit['type']);
  }
  debug('saving resource "'.$edit['name'].'" using import table : '.print_r($import_template,1),1);

  $edit['statements'] = relationship_array2statements( $edit , $subject, $import_template, NULL, $why );
  return relationship_replace_resource_statements($edit);
}




/**
 * Save all statements contained in the 'statements' array of this object or array.
 *
 * Saving the resource statements like this will wipe all old ones
 * and create new ones.
 * @see relationship_save_resource()
 */
function relationship_replace_resource_statements($edit){
  $edit = (array)$edit;
  
  if(! $edit['name']){ // This is an API, paranoia
    trigger_error('Cannot work with a term with no name identifier at all! Problem saving statements for '.print_r($edit,1),E_USER_ERROR);
    return;
  }

  // del any previous statements about this thing (short and longname to be sure)
  relationship_resource_delete_statements($edit['name']);
  if($edit['uri']){
    relationship_resource_delete_statements($edit['uri']);
  }

  foreach ($edit['statements'] as $statement) {
    relationship_save_statement($statement);
  }
  return $edit;
}



/**
 * @}
 */


/**
 * Returns whether the input string looks like a valid URI.
 *
 * Used to detect if an RDF target looks like plaintext or a reference,
 * as it's possible to have 'literals' in the second part of a triple. 
 * Not 100% guaranteed,  but a good try.
 * Literal strings CONTAINING URLs (eg comments) may confuse this
 * - Yep. Now limiting to http only - temporarily
 */
function is_uri($string) {
  $url_parts = parse_url($string);
  return (($url_parts['scheme']=='http') && $url_parts['host'])?TRUE:FALSE;
}  

/**
 * Roughly remove a bunch of non-interesting bits from long urls,  for cuter formatting
 */
function scrunch_url($url) {
  $label = preg_replace('|http://(www.)?|', '', $url);
  $label = preg_replace('|/.*[/#]|', '/...', $label);
  return $label;
}


?>
